/*
 * ============================================================================
 *  Title:    Configuration Manager
 *  Author:   J. Zbiciak
 *  $Id: cfg.h,v 1.2 1999/09/17 13:10:01 im14u2c Exp $
 * ============================================================================
 *  This module manages the machine configuration.  It does commandline
 *  parsing and processes the configuration elements that were read in 
 *  via the config-file parser.
 *
 *  CFG owns the entire machine -- it is encapsulated in a cfg_t.
 * ============================================================================
 *  CFG_INIT     -- Parse command line and get started
 *  CFG_FILE     -- Parse a config file and extend the state of the machine.
 * ============================================================================
 */

#ifndef _CFG_H_
#define _CFG_H_

/*
 * ============================================================================
 *  CFG_KEYB_T   -- Human-readable name associations for various jzIntv inputs.
 * ============================================================================
 */
typedef struct cfg_keyb_t
{
    char        *name;          /* Name of input.                           */
    char        *event;         /* Default event bound to.                  */
    v_uint_32   *word;          /* Word modified by input.                  */
    uint_32     and_mask[2];    /* Up/down AND masks.                       */
    uint_32     or_mask [2];    /* Up/down OR masks.                        */
} cfg_keyb_t;

/*
 * ============================================================================
 *  CFG_T            -- The current machine and its configuration
 * ============================================================================
 */
typedef struct cfg_t
{
    /* -------------------------------------------------------------------- */
    /*  Our baby, the Intellivision itself.                                 */
    /* -------------------------------------------------------------------- */
    periph_bus_t    intv;       /* Top level Intellivision object.          */

    /* -------------------------------------------------------------------- */
    /*  System 'peripherals' -- these aren't actual hardware devices.       */
    /* -------------------------------------------------------------------- */
    gfx_t       gfx;            /* Graphics interface layer.                */
    snd_t       snd;            /* Sound interface layer.                   */
    speed_t     speed;          /* Rate controller.                         */
    debug_t     debug;          /* Debugger hooks.                          */

    /* -------------------------------------------------------------------- */
    /*  Hardware peripherals -- these model actual pieces of the Intv.      */
    /* -------------------------------------------------------------------- */
    cp1600_t    cp1600;         /* CPU.                                     */
    ay8910_t    psg0;           /* Base unit's Prog. Sound Generator.       */
    ay8910_t    psg1;           /* ECS's Prog. Sound Generator.             */
    ivoice_t    ivoice;         /* Intellivoice.                            */
    stic_t      stic;           /* Standard Television Interface Chip       */
    pad_t       pad0;           /* Base unit's controller pads.             */
    pad_t       pad1;           /* ECS's controller pads.                   */
    mem_t       *mem;           /* Various ROMs and RAMs                    */

    /* -------------------------------------------------------------------- */
    /*  User-modifiable flags, etc.                                         */
    /* -------------------------------------------------------------------- */
    uint_32     audio_rate;     /* Sample rate for audio.  0 == no audio    */
    uint_32     psg_window;     /* Window size for PSG sliding window       */
    uint_32     full_scr;       /* Full-screen / Windowed display.          */
    uint_32     hcif_port;      /* Direct HCIF I/O address.                 */
    uint_32     debugging;      /* Debugger enabled flag.                   */
    uint_32     rate_ctl;       /* Rate control enabled flag.               */

    char        *fn_exec;       /* File name of EXEC image.                 */
    char        *fn_grom;       /* File name of GROM image.                 */
    char        *fn_game;       /* File name of GAME image.                 */
    char        *fn_gcfg;       /* File name of Game CFG file.              */

    /* -------------------------------------------------------------------- */
    /*  Key bindings                                                        */
    /* -------------------------------------------------------------------- */
    cfg_keyb_t  *bindings;
} cfg_t;


#endif
/* ======================================================================== */
/*  This program is free software; you can redistribute it and/or modify    */
/*  it under the terms of the GNU General Public License as published by    */
/*  the Free Software Foundation; either version 2 of the License, or       */
/*  (at your option) any later version.                                     */
/*                                                                          */
/*  This program is distributed in the hope that it will be useful,         */
/*  but WITHOUT ANY WARRANTY; without even the implied warranty of          */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU       */
/*  General Public License for more details.                                */
/*                                                                          */
/*  You should have received a copy of the GNU General Public License       */
/*  along with this program; if not, write to the Free Software             */
/*  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.               */
/* ======================================================================== */
/*                 Copyright (c) 1998-1999, Joseph Zbiciak                  */
/* ======================================================================== */
